# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for `gwpy.table`
"""

import os.path
import re
from io import BytesIO
from unittest import mock

import pytest

from numpy import (random, isclose, dtype, asarray, all)

import h5py

from astropy import units
from astropy.io.ascii import InconsistentTableError
from astropy.table import vstack

from ...io import ligolw as io_ligolw
from ...segments import (Segment, SegmentList)
from ...testing import utils
from ...testing.errors import pytest_skip_network_error
from ...time import LIGOTimeGPS
from ...timeseries import (TimeSeries, TimeSeriesDict)
from .. import (Table, EventTable, filters)
from ..filter import filter_table
from ..io.hacr import HACR_COLUMNS

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

TEST_DATA_DIR = os.path.join(os.path.split(__file__)[0], 'data')
TEST_XML_FILE = os.path.join(
    TEST_DATA_DIR, 'H1-LDAS_STRAIN-968654552-10.xml.gz')
TEST_OMEGA_FILE = os.path.join(TEST_DATA_DIR, 'omega.txt')


# -- mocks --------------------------------------------------------------------

def mock_hacr_connection(table, start, stop):
    """Mock a pymysql connection object to test HACR fetching
    """
    # create cursor
    cursor = mock.MagicMock()

    def execute(qstr):
        cursor._query = qstr
        return len(table)

    cursor.execute = execute
    column_regex = re.compile(r"\Aselect (.*) from ", re.I)
    select_regex = re.compile(r"where (.*) (order by .*)?\Z", re.I)

    def fetchall():
        q = cursor._query
        if "from job" in q:
            return [(1, start, stop)]
        if "from mhacr" in q:
            columns = column_regex.match(q).groups()[0].split(", ")
            selections = (
                select_regex.search(q).groups()[0].strip().split(" and ")
            )
            return map(tuple, filter_table(table, selections[3:])[columns])

    cursor.fetchall = fetchall

    # create connection
    conn = mock.MagicMock()
    conn.cursor.return_value = cursor
    return conn


# -- gwpy.table.Table (astropy.table.Table) -----------------------------------

class TestTable(object):
    TABLE = Table

    @classmethod
    def create(cls, n, names, dtypes=None):
        data = []
        for i, name in enumerate(names):
            random.seed(i)
            if dtypes:
                dtp = dtypes[i]
            else:
                dtp = None
            # use map() to support non-primitive types
            if dtype(dtp).name == 'object':
                data.append(list(map(dtp, random.rand(n) * 1000)))
            else:
                data.append((random.rand(n) * 1000).astype(dtp))
        return cls.TABLE(data, names=names)

    @pytest.fixture
    def table(self):
        return self.create(100, ['time', 'snr', 'frequency'])

    @pytest.fixture
    def emptytable(self):
        return self.create(0, ['time', 'snr', 'frequency'])

    @pytest.fixture
    def clustertable(self):
        return self.TABLE(data=[[11, 1, 1, 10, 1, 1, 9],
                               [0.0, 1.9, 1.95, 2.0, 2.05, 2.1, 4.0]],
                         names=['amplitude', 'time'])

    # -- test I/O -------------------------------

    @pytest.mark.requires("ligo.lw.lsctables")
    @pytest.mark.parametrize('ext', ['xml', 'xml.gz'])
    def test_read_write_ligolw(self, tmp_path, ext):
        table = self.create(
            100, ['peak_time', 'peak_time_ns', 'snr', 'central_freq'],
            ['i4', 'i4', 'f4', 'f4'])
        tmp = tmp_path / f"table.{ext}"

        def _read(*args, **kwargs):
            kwargs.setdefault('format', 'ligolw')
            kwargs.setdefault('tablename', 'sngl_burst')
            return self.TABLE.read(tmp, *args, **kwargs)

        def _write(*args, **kwargs):
            kwargs.setdefault('format', 'ligolw')
            kwargs.setdefault('tablename', 'sngl_burst')
            return table.write(tmp, *args, **kwargs)

        # check simple write (using open file descriptor, not file path)
        with tmp.open("w+b") as f:
            table.write(f, format='ligolw', tablename='sngl_burst')

        # check simple read
        t2 = _read()
        utils.assert_table_equal(table, t2, almost_equal=True)
        assert t2.meta.get('tablename', None) == 'sngl_burst'

        # check numpy type casting works
        from ligo.lw.lsctables import LIGOTimeGPS as LigolwGPS
        t3 = _read(columns=['peak'])
        assert len(t3) == 100
        assert isinstance(t3['peak'][0], LigolwGPS)
        t3 = _read(columns=['peak'], use_numpy_dtypes=True)
        assert len(t3) == 100
        assert t3['peak'].dtype == dtype('float64')
        utils.assert_array_equal(
            t3['peak'], table['peak_time'] + table['peak_time_ns'] * 1e-9)

        # check reading multiple tables works
        t3 = self.TABLE.read([tmp, tmp], format='ligolw',
                             tablename='sngl_burst')
        utils.assert_table_equal(vstack((t2, t2)), t3)

        # check writing to existing file raises IOError
        with pytest.raises(
            IOError,
            match=f"^File exists: {tmp}$",
        ):
            _write()

        # check overwrite=True, append=False rewrites table
        _write(overwrite=True)
        t3 = _read()
        utils.assert_table_equal(t2, t3)

        # check append=True duplicates table
        _write(append=True)
        t3 = _read()
        utils.assert_table_equal(vstack((t2, t2)), t3)

        # check overwrite=True, append=True rewrites table
        _write(append=True, overwrite=True)
        t3 = _read()
        utils.assert_table_equal(t2, t3)

        # write another table and check we can still get back the first
        insp = self.create(10, ['end_time', 'snr', 'chisq_dof'])
        insp.write(tmp, format='ligolw', tablename='sngl_inspiral',
                   append=True)
        t3 = _read()
        utils.assert_table_equal(t2, t3)

        # write another table with append=False and check the first table
        # is gone
        insp.write(tmp, format='ligolw', tablename='sngl_inspiral',
                   append=False, overwrite=True)
        with pytest.raises(
            ValueError,
            match="^document must contain exactly one sngl_burst table$",
        ):
            _read()

    @pytest.mark.requires("ligo.lw.lsctables")
    def test_read_write_ligolw_property_columns(self, tmp_path):
        table = self.create(100, ['peak', 'snr', 'central_freq'],
                            ['f8', 'f4', 'f4'])

        # write table
        tmp = tmp_path / "test.xml"
        table.write(tmp, format='ligolw', tablename='sngl_burst')

        # read raw ligolw and check gpsproperty was unpacked properly
        llw = io_ligolw.read_table(tmp, tablename='sngl_burst')
        for col in ('peak_time', 'peak_time_ns'):
            assert col in llw.columnnames
        with io_ligolw.patch_ligotimegps():
            utils.assert_array_equal(
                asarray([row.peak for row in llw]),
                table['peak'],
            )

        # read table and assert gpsproperty was repacked properly
        t2 = self.TABLE.read(
            tmp,
            columns=table.colnames,
            use_numpy_dtypes=True,
        )
        utils.assert_table_equal(t2, table, almost_equal=True)

    @pytest.mark.requires("ligo.lw.lsctables")
    def test_read_ligolw_get_as_exclude(self, tmp_path):
        table = self.TABLE(
            rows=[
                ("H1", 0.0, 4, 0),
                ("L1", 0.62831, 4, 0),
                ("V1", 0.31415, 4, 0),
            ],
            names=("instrument", "offset", "process_id", "time_slide_id"),
        )

        # write a file
        tmp = tmp_path / "test.xml"
        table.write(
            tmp,
            format="ligolw",
            tablename="time_slide",
        )

        # read it back and assert that we the `instrument` table is
        # read properly
        t2 = table.read(
            tmp,
            tablename="time_slide",
        )
        t2.sort("instrument")
        utils.assert_table_equal(t2, table)

    @pytest.mark.requires("uproot")
    def test_read_write_root(self, table, tmp_path):
        tmp = tmp_path / "table.root"

        # check write
        table.write(tmp)

        # check read gives back same table
        t2 = self.TABLE.read(tmp)
        utils.assert_table_equal(table, t2)

        # test selections work
        segs = SegmentList([Segment(100, 200), Segment(400, 500)])
        t2 = self.TABLE.read(
            tmp,
            selection=['200 < frequency < 500',
                       ('time', filters.in_segmentlist, segs)],
        )
        utils.assert_table_equal(
            t2,
            filter_table(
                table,
                'frequency > 200',
                'frequency < 500',
                ('time', filters.in_segmentlist, segs),
            ),
        )

    @pytest.mark.requires("uproot")
    def test_write_root_overwrite(self, table, tmp_path):
        tmp = tmp_path / "table.root"
        table.write(tmp)

        # assert failure with overwrite=False (default)
        with pytest.raises(OSError):
            table.write(tmp)

        # assert works with overwrite=True
        table.write(tmp, overwrite=True)

    @pytest.mark.requires("uproot")
    def test_read_root_multiple_trees(self, tmp_path):
        uproot = pytest.importorskip("uproot")
        tmp = tmp_path / "table.root"
        with uproot.create(tmp) as root:
            a = root.mktree("a", {"branch": "int32"})
            a.extend({"branch": asarray([1, 2, 3, 4, 5])})
            root.mktree("b", {"branch": "int32"})
        with pytest.raises(
            ValueError,
            match="^Multiple trees found",
        ):
            self.TABLE.read(tmp)
        self.TABLE.read(tmp, treename="a")

    @pytest.mark.requires("uproot")
    def test_read_write_root_append(self, table, tmp_path):
        tmp = tmp_path / "table.root"
        # write one tree
        table.write(tmp, treename="a")
        # write a second tree
        table.write(tmp, treename="b", append=True)
        # check that we can read both trees
        self.TABLE.read(tmp, treename="a")
        self.TABLE.read(tmp, treename="b")

    @pytest.mark.requires("uproot")
    def test_read_write_root_append_overwrite(self, table, tmp_path):
        tmp = tmp_path / "table.root"
        # write one tree
        table.write(tmp, treename="a")
        # write a second tree
        table.write(tmp, treename="b", append=True)
        # overwrite the first tree
        t2 = table[:50]
        t2.write(tmp, treename="a", overwrite=True, append=True)
        # check that we can read the original 'b' tree and the new
        # 'a' tree
        utils.assert_table_equal(table, self.TABLE.read(tmp, treename="b"))
        utils.assert_table_equal(t2, self.TABLE.read(tmp, treename="a"))

    @pytest.mark.requires("LDAStools.frameCPP")
    def test_read_write_gwf(self, tmp_path):
        table = self.create(100, ['time', 'blah', 'frequency'])
        columns = table.dtype.names
        tmp = tmp_path / "table.gwf"

        # check write
        try:
            table.write(tmp, 'test_read_write_gwf')
        except TypeError as exc:  # pragma: no-cover
            msg = str(exc).splitlines()[0]
            for err, ref in [
                ("ParamList_type", 57),
                ("Unable to translate parameter to Parameters_type", 146),
            ]:
                if err in msg:
                    pytest.skip(
                        f"bug in python-ldas-tools-framecpp: '{msg}', "
                        "see https://git.ligo.org/computing/ldastools/"
                        f"LDAS_Tools/-/issues/{ref}",
                    )
            raise

        # check read gives back same table
        t2 = self.TABLE.read(tmp, 'test_read_write_gwf', columns=columns)
        utils.assert_table_equal(table, t2, meta=False, almost_equal=True)

        # check selections works
        t3 = self.TABLE.read(tmp, 'test_read_write_gwf',
                             columns=columns, selection='frequency>500')
        utils.assert_table_equal(
            filter_table(t2, 'frequency>500'), t3)


class TestEventTable(TestTable):
    TABLE = EventTable

    def test_get_time_column(self, table):
        """Check that `_get_time_colum` works on name
        """
        assert table._get_time_column() == "time"

    def test_get_time_column_case(self, table):
        """Check that `_get_time_colum` works on name case-insensitively
        """
        table.rename_column("time", "TiMe")
        assert table._get_time_column() == 'TiMe'

    def test_get_time_column_gps_type(self):
        """Check that `_get_time_column` works on dtype
        """
        # check that single GPS column can be identified
        t = self.create(1, ("a", "b"), dtypes=(float, LIGOTimeGPS))
        assert t._get_time_column() == "b"

    def test_get_time_column_error_no_match(self, table):
        """Check that `_get_time_column` raises the right exception
        when no matches are found
        """
        t = self.create(1, ("a",))
        with pytest.raises(
            ValueError,
            match=" no columns named 'time'",
        ):
            t._get_time_column()

    def test_get_time_column_error_multiple_match(self):
        # check that two GPS columns causes issues
        t = self.create(
            10,
            ("a", "b", "c"),
            dtypes=(float, LIGOTimeGPS, LIGOTimeGPS),
        )
        with pytest.raises(
            ValueError,
            match=" multiple columns named 'time'",
        ):
            t._get_time_column()

    def test_get_time_column_error_empty(self):
        """Check that `_get_time_column` errors properly on an empty table
        """
        t = self.create(0, ("a",))
        with pytest.raises(ValueError):
            t._get_time_column()

    def test_filter(self, table):
        """Test that `EventTable.filter` works with a simple filter statement
        """
        # check simple filter
        lowf = table.filter('frequency < 100')
        assert isinstance(lowf, type(table))
        assert len(lowf) == 11
        assert isclose(lowf['frequency'].max(), 96.5309156606)

    def test_filter_empty(self, table):
        """Test that `EventTable.filter` works with an empty table
        """
        assert len(table.filter('snr>5', 'snr<=5')) == 0

    def test_filter_chaining(self, table):
        """Test that chaining filters works with `EventTable.filter`
        """
        loud = table.filter('snr > 100')
        lowf = table.filter('frequency < 100')
        lowfloud = table.filter('frequency < 100', 'snr > 100')
        brute = type(table)(
            rows=[tuple(row) for row in lowf if row in loud],
            names=table.dtype.names,
        )
        utils.assert_table_equal(brute, lowfloud)

    def test_filter_range(self, table):
        """Test that `EventTable.filter` works with a range statement
        """
        # check double-ended filter
        midf = table.filter('100 < frequency < 1000')
        utils.assert_table_equal(
            midf,
            table.filter('frequency > 100').filter('frequency < 1000'),
        )

    def test_filter_function(self, table):
        """Test that `EventTable.filter` works with a filter function
        """
        def my_filter(column, threshold):
            return column < threshold

        lowf = table.filter(("frequency", my_filter, 100))
        assert len(lowf) == 11

    def test_filter_function_multiple(self, table):
        """Test that `EventTable.filter` works with a filter function
        that requires multiple columns
        """
        def my_filter(table, threshold):
            return table["snr"] * table["frequency"] > threshold

        filtered = table.filter((("snr", "frequency"), my_filter, 100000))
        assert len(filtered) == 64

    def test_filter_in_segmentlist(self, table):
        """Test `EventTable.filter` with `in_segmentlist`
        """
        # check filtering on segments works
        segs = SegmentList([Segment(100, 200), Segment(400, 500)])
        inseg = table.filter(('time', filters.in_segmentlist, segs))
        brute = type(table)(
            rows=[tuple(row) for row in table if row['time'] in segs],
            names=table.colnames,
        )
        utils.assert_table_equal(inseg, brute)

    def test_filter_in_segmentlist_empty(self, table):
        """Test `EventTable.filter` with `in_segmentlist` and an empty table
        """
        # check empty segmentlist is handled well
        utils.assert_table_equal(
            table.filter(('time', filters.in_segmentlist, SegmentList())),
            type(table)(names=table.colnames),
        )

    def test_filter_not_in_segmentlist(self, table):
        """Test `EventTable.filter` with `not_in_segmentlist`
        """
        segs = SegmentList([Segment(100, 200), Segment(400, 500)])
        notsegs = SegmentList([Segment(0, 1000)]) - segs
        inseg = table.filter(('time', filters.in_segmentlist, segs))
        utils.assert_table_equal(
            inseg, table.filter(('time', filters.not_in_segmentlist, notsegs)),
        )
        utils.assert_table_equal(
            table,
            table.filter(('time', filters.not_in_segmentlist, SegmentList())),
        )

    def test_event_rates(self, table):
        """Test :meth:`gwpy.table.EventTable.event_rate`
        """
        rate = table.event_rate(1)
        assert isinstance(rate, TimeSeries)
        assert rate.sample_rate == 1 * units.Hz

    @pytest.mark.requires("lal")
    def test_event_rates_gpsobject(self, table):
        """Test that `EventTable.event_rate` can handle object dtypes
        """
        rate = table.event_rate(1)

        from lal import LIGOTimeGPS as LalGps
        lgps = list(map(LalGps, table['time']))
        t2 = type(table)(data=[lgps], names=['time'])
        rate2 = t2.event_rate(1, start=table['time'].min())

        utils.assert_quantity_sub_equal(rate, rate2)

    def test_event_rates_start_end(self):
        """Check that `EventTable.event_rate` can function without explicit
        time column (and no data) if and only if start/end are both given.
        """
        t2 = self.create(10, names=['a', 'b'])
        with pytest.raises(
            ValueError,
            match="please give `timecolumn` keyword",
        ):
            t2.event_rate(1)
        with pytest.raises(ValueError):
            t2.event_rate(1, start=0)
        with pytest.raises(ValueError):
            t2.event_rate(1, end=1)
        t2.event_rate(1, start=0, end=10)

    def test_binned_event_rates(self, table):
        rates = table.binned_event_rates(100, 'snr', [10, 100],
                                         timecolumn='time')
        assert isinstance(rates, TimeSeriesDict)
        assert list(rates.keys()), [10, 100]
        assert rates[10].max() == 0.14 * units.Hz
        assert rates[10].name == 'snr >= 10'
        assert rates[100].max() == 0.13 * units.Hz
        assert rates[100].name == 'snr >= 100'
        table.binned_event_rates(100, 'snr', [10, 100], operator='in')
        table.binned_event_rates(100, 'snr', [(0, 10), (10, 100)])

        # check that method can function without explicit time column
        # (and no data) if and only if start/end are both given
        t2 = self.create(0, names=['a', 'b'])
        with pytest.raises(
            ValueError,
            match="please give `timecolumn` keyword",
        ):
            t2.binned_event_rates(1, 'a', (10, 100))
        with pytest.raises(ValueError):
            t2.binned_event_rates(1, 'a', (10, 100), start=0)
        with pytest.raises(ValueError):
            t2.binned_event_rates(1, 'a', (10, 100), end=1)
        t2.binned_event_rates(1, 'a', (10, 100), start=0, end=10)

    def test_plot(self, table):
        with pytest.deprecated_call():
            plot = table.plot('time', 'frequency', color='snr')
            plot.close()

    def test_scatter(self, table):
        plot = table.scatter('time', 'frequency', color='snr')
        plot.save(BytesIO(), format='png')
        plot.close()

    def test_hist(self, table):
        plot = table.hist('snr')
        assert len(plot.gca().patches) == 10
        plot.save(BytesIO(), format='png')
        plot.close()

    def test_get_column(self, table):
        utils.assert_array_equal(table.get_column('snr'), table['snr'])

    def test_cluster(self, clustertable):
        # check that the central data points are all clustered away,
        # the original table is unchanged, and all points return their
        # intended values
        t = clustertable.cluster('time', 'amplitude', 0.6)
        assert len(t) == 3
        assert len(clustertable) == 7
        assert all(t['amplitude'] == [11, 10, 9])
        assert all(t['time'] == [0.0, 2.0, 4.0])

    def test_single_point_cluster(self, clustertable):
        # check that a large cluster window returns at least one data point
        t = clustertable.cluster('time', 'amplitude', 10)
        assert len(t) == 1
        assert all(t['amplitude'] == [11])
        assert all(t['time'] == [0.0])

    def test_cluster_window(self, clustertable):
        # check that a non-positive window throws an appropriate ValueError
        with pytest.raises(
            ValueError,
            match="^Window must be a positive value$",
        ):
            clustertable.cluster('time', 'amplitude', 0)

    def test_cluster_multiple(self, clustertable):
        # check that after clustering a table, clustering the table a
        # second time with the same parameters returns the same result
        t_clustered = clustertable.cluster('time', 'amplitude', 0.6)
        utils.assert_table_equal(
            t_clustered,
            t_clustered.cluster('time', 'amplitude', 0.6),
        )

    def test_cluster_empty(self, emptytable):
        # check that clustering an empty table is a no-op
        t = emptytable.cluster('time', 'amplitude', 0.6)
        utils.assert_table_equal(t, emptytable)

    # -- test I/O -------------------------------

    def test_read_write_hdf5(self, table, tmp_path):
        # check that our overrides of astropy's H5 reader
        # didn't break everything
        tmp = tmp_path / "table.h5"
        table.write(tmp, path="/data")
        t2 = self.TABLE.read(tmp, path="/data")
        utils.assert_table_equal(t2, table)

        t2 = self.TABLE.read(
            tmp,
            path="/data",
            selection="frequency>500",
            columns=["time", "snr"],
        )
        utils.assert_table_equal(
            t2,
            filter_table(table, "frequency>500")[("time", "snr")],
        )

    @pytest.mark.parametrize('fmtname', ('Omega', 'cWB'))
    def test_read_write_ascii(self, table, tmp_path, fmtname):
        fmt = 'ascii.{}'.format(fmtname.lower())
        tmp = tmp_path / "table.txt"

        # check write/read returns the same table
        with tmp.open("w") as fobj:
            table.write(fobj, format=fmt)
        t2 = self.TABLE.read(tmp, format=fmt)
        utils.assert_table_equal(table, t2, almost_equal=True)

        # check that we can use selections and column filtering
        t2 = self.TABLE.read(
            tmp,
            format=fmt,
            selection="frequency>500",
            columns=["time", "snr"],
        )
        utils.assert_table_equal(
            t2,
            filter_table(table, "frequency>500")[("time", "snr")],
            almost_equal=True,
        )

    @pytest.mark.parametrize('fmtname', ('Omega', 'cWB'))
    def test_read_write_ascii_error(self, table, tmp_path, fmtname):
        tmp = tmp_path / "table.txt"
        with tmp.open("w"):
            pass  # write empty file

        # assert reading blank file doesn't work with column name error
        with pytest.raises(
            InconsistentTableError,
            match=f"^No column names found in {fmtname} header$",
        ):
            self.TABLE.read(
                tmp,
                format="ascii.{}".format(fmtname.lower()),
            )

    @pytest.fixture
    def snaxtable(self):
        channel = 'H1:FAKE'
        table = self.create(
            100,
            names=[
                'time',
                'snr',
                'frequency',
            ],
        )
        table["channel"] = channel
        return table

    @staticmethod
    @pytest.fixture
    def snaxfile(snaxtable, tmp_path):
        tmp = tmp_path / "SNAX-0-0.h5"
        channel = snaxtable[0]["channel"]
        tmptable = snaxtable.copy()
        tmptable.remove_column("channel")
        with h5py.File(tmp, "w") as h5f:
            group = h5f.create_group(channel)
            group.create_dataset(data=tmptable, name='0.0_20.0')
        return tmp

    def test_read_snax(self, snaxtable, snaxfile):
        """Check that we can read a SNAX-format HDF5 file
        """
        table = self.TABLE.read(snaxfile, format='hdf5.snax')
        utils.assert_table_equal(snaxtable, table)

    def test_read_snax_channel(self, snaxtable, snaxfile):
        """Check that we can read a SNAX-format HDF5 file specifying
        the channel
        """
        table = self.TABLE.read(
            snaxfile,
            format='hdf5.snax',
            channels="H1:FAKE",
        )
        utils.assert_table_equal(snaxtable, table)

    def test_read_snax_selection_columns(self, snaxtable, snaxfile):
        """Check that the selection and columns kwargs work when
        reading from a SNAX-format file
        """
        # test with selection and columns
        table = self.TABLE.read(
            snaxfile,
            channels="H1:FAKE",
            format='hdf5.snax',
            selection='snr>.5',
            columns=('time', 'snr'),
        )
        utils.assert_table_equal(
            table,
            filter_table(snaxtable, 'snr>.5')[('time', 'snr')],
        )

    def test_read_snax_compact(self, snaxtable, snaxfile):
        """Check that the selection and columns kwargs work when
        reading from a SNAX-format file
        """
        # test compact representation of channel column
        t2 = self.TABLE.read(snaxfile, compact=True, format='hdf5.snax')

        # group by channel and drop channel column
        tables = {}
        t2 = t2.group_by('channel')
        t2.remove_column('channel')
        for key, group in zip(t2.groups.keys, t2.groups):
            channel = t2.meta['channel_map'][key['channel']]
            tables[channel] = self.TABLE(group, copy=True)

        # verify table groups are identical
        t_ref = snaxtable.copy().group_by('channel')
        t_ref.remove_column('channel')
        for key, group in zip(t_ref.groups.keys, t_ref.groups):
            channel = key['channel']
            utils.assert_table_equal(group, tables[channel])

    def test_read_snax_errors(self, snaxtable, snaxfile):
        """Check error handling when reading from a SNAX-format file
        """
        missing = ["H1:FAKE", 'H1:MISSING']
        with pytest.raises(ValueError):
            self.TABLE.read(snaxfile, channels=missing, format='hdf5.snax')

        with pytest.warns(UserWarning):
            table = self.TABLE.read(
                snaxfile,
                channels=missing,
                format='hdf5.snax',
                on_missing='warn',
            )
        utils.assert_table_equal(snaxtable, table)

    @pytest.fixture(scope="module")
    def hacr_table(self):
        """Create a table of HACR-like data, and patch
        `pymysql.connect` to return it
        """
        table = self.create(100, names=HACR_COLUMNS)
        connect = mock.patch(
            "pymysql.connect",
            return_value=mock_hacr_connection(table, 123, 456),
        )
        try:
            connect.start()
        except ImportError as exc:  # pragma: no-cover
            pytest.skip(str(exc))
        yield table
        connect.stop()

    def test_fetch_hacr(self, hacr_table):
        t2 = self.TABLE.fetch('hacr', 'X1:TEST-CHANNEL', 123, 456)

        # check type matches
        assert type(t2) is self.TABLE

        # check response is correct
        utils.assert_table_equal(hacr_table, t2)

    def test_fetch_hacr_columns(self, hacr_table):
        t2 = self.TABLE.fetch('hacr', 'X1:TEST-CHANNEL', 123, 456,
                              columns=['gps_start', 'snr'],
                              selection='freq_central>500')
        utils.assert_table_equal(
            filter_table(hacr_table, 'freq_central>500')['gps_start', 'snr'],
            t2,
        )

    @pytest_skip_network_error
    def test_fetch_open_data(self):
        table = self.TABLE.fetch_open_data("GWTC-1-confident")
        assert len(table)
        assert {
            "mass_1_source",
            "luminosity_distance",
            "chi_eff"
        }.intersection(table.colnames)
        # check unit parsing worked
        assert table["luminosity_distance"].unit == "Mpc"

    @pytest_skip_network_error
    def test_fetch_open_data_kwargs(self):
        table = self.TABLE.fetch_open_data(
            "GWTC-1-confident",
            selection="mass_1_source < 5",
            columns=[
                "name",
                "mass_1_source",
                "mass_2_source",
                "luminosity_distance"
            ],
        )
        assert len(table) == 1
        assert table[0]["name"] == "GW170817-v3"
        assert set(table.colnames) == {
            "name",
            "mass_1_source",
            "mass_2_source",
            "luminosity_distance"
        }



# -*- coding: utf-8 -*-
# Copyright (C) California Institute of Technology (2019-2022)
#               Pensylvania State University (2019)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from the GstLAL online GW search
"""

import re
from os.path import basename

from astropy.table import join

from ...io.registry import (register_reader, register_identifier)
from ...io.ligolw import is_ligolw
from .ligolw import read_table
from .. import EventTable
from .pycbc import get_mchirp

__author__ = 'Derk Davis <derek.davis@ligo.org>'
__credits__ = 'Patrick Godwin <patrick.godwin@ligo.org>'

GSTLAL_FORMAT = 'ligolw.gstlal'
GSTLAL_SNGL_FORMAT = 'ligolw.gstlal.sngl'
GSTLAL_COINC_FORMAT = 'ligolw.gstlal.coinc'

GSTLAL_FILENAME = re.compile('([A-Z][0-9])+-LLOID-[0-9.]+-[0-9.]+.xml.gz')


# singles format
def read_gstlal_sngl(source, **kwargs):
    """Read a `sngl_inspiral` table from one or more GstLAL LIGO_LW XML files

    source : `file`, `str`, :class:`~ligo.lw.ligolw.Document`, `list`
        one or more open files, file paths, or LIGO_LW `Document` objects

    **kwargs
        keyword arguments for the read, or conversion functions

    See also
    --------
    gwpy.io.ligolw.read_table
        for details of keyword arguments for the read operation
    gwpy.table.io.ligolw.to_astropy_table
        for details of keyword arguments for the conversion operation
    """
    from ligo.lw import lsctables
    extra_cols = []
    derived_cols = []
    val_col = lsctables.TableByName['sngl_inspiral'].validcolumns
    if 'columns' in kwargs:
        for name in kwargs['columns'].copy():
            if name in GET_COLUMN:
                derived_cols.append(name)
                kwargs['columns'].remove(name)
                required_cols = GET_COLUMN_EXTRA[name]
                missing_cols = [c for c in required_cols
                                if c not in kwargs['columns']]
                for r_col in missing_cols:
                    kwargs['columns'].append(r_col)
                    extra_cols.append(r_col)
            elif name not in val_col:
                name_list = list(val_col.keys())+list(GET_COLUMN.keys())
                raise ValueError(f"'{name}' is not a valid column name. "
                                 f"Valid column names: {name_list}")
    events = read_table(source, tablename='sngl_inspiral', **kwargs)
    for col_name in derived_cols:
        col_data = GET_COLUMN[col_name](events)
        events.add_column(col_data, name=col_name)
    for col_name in extra_cols:
        events.remove_column(col_name)
    return events


# coinc format
def read_gstlal_coinc(source, **kwargs):
    """Read a `Table` containing coincident event information
    from one or more GstLAL LIGO_LW XML files

    source : `file`, `str`, :class:`~ligo.lw.ligolw.Document`, `list`
        one or more open files, file paths, or LIGO_LW `Document` objects

    **kwargs
        keyword arguments for the read, or conversion functions

    See also
    --------
    gwpy.io.ligolw.read_table
        for details of keyword arguments for the read operation
    gwpy.table.io.ligolw.to_astropy_table
        for details of keyword arguments for the conversion operation
    """
    from ligo.lw import lsctables
    extra_cols = []
    if 'columns' in kwargs:
        columns = kwargs['columns']
        kwargs.pop('columns')
        val_col_inspiral = lsctables.TableByName['coinc_inspiral'].validcolumns
        val_col_event = lsctables.TableByName['coinc_event'].validcolumns
        for name in columns:
            if (name not in val_col_inspiral) and (name not in val_col_event):
                name_list = list(val_col_inspiral.keys()) + \
                    list(val_col_event.keys())
                raise ValueError(f"'{name}' is not a valid column name. "
                                 f"Valid column names: {name_list}")
        if 'coinc_event_id' not in columns:
            columns.append('coinc_event_id')
            extra_cols.append('coinc_event_id')
        inspiral_cols = [col for col in columns if col in val_col_inspiral]
        event_cols = [col for col in columns if col in val_col_event]
        inspiral_cols.append('coinc_event_id')
        coinc_inspiral = read_table(source, tablename='coinc_inspiral',
                                    columns=inspiral_cols, **kwargs)
        coinc_event = read_table(source, tablename='coinc_event',
                                 columns=event_cols, **kwargs)
    else:
        coinc_inspiral = read_table(source, tablename='coinc_inspiral',
                                    **kwargs)
        coinc_event = read_table(source, tablename='coinc_event', **kwargs)
    events = join(coinc_inspiral, coinc_event, keys="coinc_event_id",
                  metadata_conflicts='silent')
    events.meta['tablename'] = 'gstlal_coinc_inspiral'
    for col_name in extra_cols:
        events.remove_column(col_name)
    return events


# combined format
def read_gstlal(source, triggers='sngl', **kwargs):
    """Read a `Table` from one or more GstLAL LIGO_LW XML files

    source : `file`, `str`, :class:`~ligo.lw.ligolw.Document`, `list`
        one or more open files, file paths, or LIGO_LW `Document` objects

    triggers : `str`, optional
        the `Name` of the relevant `Table` to read, if not given a table will
        be returned if only one exists in the document(s).
        'sngl' for single-detector trigger information,
        'coinc' for coincident trigger information

    **kwargs
        keyword arguments for the read, or conversion functions

    See also
    --------
    gwpy.io.ligolw.read_table
        for details of keyword arguments for the read operation
    gwpy.table.io.ligolw.to_astropy_table
        for details of keyword arguments for the conversion operation
    """

    if triggers == 'sngl':
        return read_gstlal_sngl(source, **kwargs)
    if triggers == 'coinc':
        return read_gstlal_coinc(source, **kwargs)
    else:
        raise ValueError("The 'triggers' argument must be 'sngl' or 'coinc'")


def identify_gstlal(origin, filepath, fileobj, *args, **kwargs):
    """Identify a GstLAL file as a ligolw file with the correct name
    """
    if is_ligolw(origin, filepath, fileobj, *args, **kwargs) and (
            filepath is not None
            and GSTLAL_FILENAME.match(basename(filepath))):
        return True
    return False


# registers for unified I/O
register_identifier(GSTLAL_FORMAT, EventTable, identify_gstlal)
register_reader(GSTLAL_SNGL_FORMAT, EventTable, read_gstlal_sngl)
register_reader(GSTLAL_COINC_FORMAT, EventTable, read_gstlal_coinc)
register_reader(GSTLAL_FORMAT, EventTable, read_gstlal)

# -- processed columns --------------------------------------------------------
#
# Here we define methods required to build commonly desired columns that
# are just a combination of the basic columns.
#
# Each method should take in a `~gwpy.table.Table` and return a `numpy.ndarray`

GET_COLUMN = {}
GET_COLUMN_EXTRA = {}


def get_snr_chi(events, snr_pow=2., chi_pow=2.):
    """Calculate the 'SNR chi' column for this GstLAL ligolw table group
    """
    snr = events['snr'][:]
    chisq = events['chisq'][:]
    snr_chi = snr**snr_pow / chisq**(chi_pow/2.)
    return snr_chi


GET_COLUMN['snr_chi'] = get_snr_chi
GET_COLUMN_EXTRA['snr_chi'] = ['snr', 'chisq']


def get_chi_snr(events, snr_pow=2., chi_pow=2.):
    """Calculate the 'chi SNR' column for this GstLAL ligolw table group,
    reciprocal of the 'SNR chi' column
    """
    return 1./get_snr_chi(events, snr_pow, chi_pow)


GET_COLUMN['chi_snr'] = get_chi_snr
GET_COLUMN_EXTRA['chi_snr'] = ['snr', 'chisq']

# use same function as pycbc
GET_COLUMN['mchirp'] = get_mchirp
GET_COLUMN_EXTRA['mchirp'] = ['mass1', 'mass2']



# -*- coding: utf-8 -*-
# Copyright (C) Evan Goetz (2021)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Transfer function plots
"""

from astropy.time import Time
from collections import OrderedDict
import numpy as np

from ..plot.bode import BodePlot
from ..plot.tex import label_to_latex
from .cliproduct import (TransferFunctionProduct, FFTMixin)
from ..plot.gps import GPS_SCALES

__author__ = 'Evan Goetz <evan.goetz@ligo.org>'


class TransferFunction(FFTMixin, TransferFunctionProduct):
    """Plot transfer function between a reference time series and one
    or more other time series
    """
    action = 'transferfunction'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ref_chan = self.args.ref or self.chan_list[0]
        # deal with channel type appendages
        if ',' in self.ref_chan:
            self.ref_chan = self.ref_chan.split(',')[0]
        self.plot_dB = self.args.plot_dB
        self.subplot = None
        self.test_chan = self.chan_list[1]
        self.tfs = []

    @property
    def ax(self):  # pylint: disable=invalid-name
        """The current `~matplotlib.axes.Axes` of this product's plot
        """
        return self.plot.axes[self.subplot]

    @classmethod
    def arg_channels(cls, parser):
        group = super().arg_channels(parser)
        group.add_argument('--ref', help='Reference channel against which '
                                         'others will be compared')
        return group

    @classmethod
    def arg_yaxis(cls, parser):
        group = cls._arg_axis('ymag', parser, scale='log')
        group.add_argument('--plot-dB', action='store_true',
                           help='Plot transfer function in dB')
        group = cls._arg_axis('yphase', parser, scale='linear')

        return group

    def get_title(self):
        gps = self.start_list[0]
        utc = Time(gps, format='gps', scale='utc').iso
        tstr = f'{utc} | {gps} ({self.duration})'

        fftstr = f'fftlength={self.args.secpfft}, overlap={self.args.overlap}'

        return ', '.join([tstr, fftstr])

    def _finalize_arguments(self, args):
        if args.ymagscale is None:
            args.ymagscale = 'log'
        if args.plot_dB:
            args.ymagscale = 'linear'

        return super()._finalize_arguments(args)

    def get_ylabel(self):
        """Text for y-axis label
        """
        ylabelstr = ''
        if self.subplot == 0:
            ylabelstr = 'Magnitude'
            if self.plot_dB:
                ylabelstr += ' [dB]'
        if self.subplot == 1:
            ylabelstr = 'Phase [deg.]'

        return ylabelstr

    def get_suptitle(self):
        """Start of default super title, first channel is appended to it
        """
        return f"Transfer function: {self.test_chan}/{self.ref_chan}"

    def set_axes_properties(self):

        for subplot in [0, 1]:
            self.subplot = subplot
            self.scale_axes_from_data()
            self.set_xaxis_properties()
            self.set_yaxis_properties()

    def _set_axis_properties(self, axis):
        """Generic method to set properties for X/Y axis
        on a specific subplot
        """
        def _get(param):
            ret = getattr(self.plot.axes[self.subplot],
                          f'get_{axis[0].lower()}{param}')()
            return ret

        def _set(param, *args, **kwargs):
            if axis.lower().startswith('y'):
                ret = getattr(self.plot.axes[self.subplot], f'set_y{param}')(
                    *args, **kwargs)
            else:
                ret = getattr(self.plot.axes[self.subplot], f'set_x{param}')(
                    *args, **kwargs)
            return ret

        scale = getattr(self.args, f'{axis}scale')
        label = getattr(self.args, f'{axis}label')
        min_ = getattr(self.args, f'{axis}min')
        max_ = getattr(self.args, f'{axis}max')

        # parse limits
        if (
            scale == 'auto-gps'
            and min_ is not None
            and max_ is not None
            and max_ < 1e8
        ):
            limits = (min_, min_ + max_)
        else:
            limits = (min_, max_)

        # set limits
        if limits[0] is not None or limits[1] is not None:
            _set('lim', *limits)

        # set scale
        if scale:
            _set('scale', scale)

        # reset scale with epoch if using GPS scale
        if _get('scale') in GPS_SCALES:
            _set('scale', scale, epoch=self.args.epoch)

        # set label
        if label is None:
            if axis.lower().startswith('y'):
                label = getattr(self, 'get_ylabel')()
            else:
                label = getattr(self, 'get_xlabel')()
        if self.subplot == 0 and axis == 'x':
            label = None
        if label:
            if self.usetex:
                label = label_to_latex(label)
            _set('label', label)

        # log
        limits = _get('lim')
        scale = _get('scale')
        label = _get('label')
        self.log(
            2,
            f'{axis.upper()}-axis parameters | '
            f'scale: {scale} | '
            f'limits: {limits[0]!s} - {limits[1]!s}'
        )
        self.log(3, (f'{axis.upper()}-axis label: {label}'))

    def set_yaxis_properties(self):
        """Set properties for Y-axis
        """
        if self.subplot == 0:
            self._set_axis_properties('ymag')
        else:
            self._set_axis_properties('yphase')

    def scale_axes_from_data(self):
        """Restrict data limits for Y-axis based on what you can see
        """
        # get tight limits for X-axis
        if self.args.xmin is None:
            self.args.xmin = min(tf.xspan[0] for tf in self.tfs)
            # this is then typically zero, so if the xscale is log or None
            # we'll need to set to be the next bin higher (one step of df)
            if (self.args.xmin == 0
                    and (self.args.xscale == 'log'
                         or self.args.xscale is None)):
                self.args.xmin = min(tf.df.value for tf in self.tfs)
        if self.args.xmax is None:
            self.args.xmax = max(tf.xspan[1] for tf in self.tfs)

        # autoscale view for Y-axis
        cropped = [tf.crop(self.args.xmin, self.args.xmax) for
                   tf in self.tfs]
        ymin = None
        ymax = None
        for tf in cropped:
            if self.subplot == 0:
                if self.plot_dB:
                    vals = 20 * np.log10(abs(tf.value))
                else:
                    vals = abs(tf.value)
            else:
                vals = np.angle(tf.value, deg=True)
            minval = min(vals)
            maxval = max(vals)
            if ymin is None or minval < ymin:
                ymin = minval
            if ymax is None or maxval > ymax:
                ymax = maxval
        self.ax.yaxis.set_data_interval(ymin, ymax, ignore=True)
        self.ax.autoscale_view(scalex=False)

    def set_plot_properties(self):
        """Finalize figure object and show() or save()
        """
        self.set_axes_properties()
        self.subplot = 0
        self.set_title(self.args.title)
        self.set_suptitle(self.args.suptitle)
        self.set_grid(not self.args.nogrid)
        self.subplot = 1
        self.set_grid(not self.args.nogrid)

    def make_plot(self):
        """Generate the transfer function plot from the time series
        """
        args = self.args

        fftlength = float(args.secpfft)
        overlap = args.overlap
        self.log(2, "Calculating transfer function secpfft: "
                 f"{fftlength}, overlap: {overlap}")
        if overlap is not None:
            overlap *= fftlength

        self.log(3, f"Reference channel: {self.ref_chan}")

        # group data by segment
        groups = OrderedDict()
        for series in self.timeseries:
            seg = series.span
            try:
                groups[seg][series.channel.name] = series
            except KeyError:
                groups[seg] = OrderedDict()
                groups[seg][series.channel.name] = series

        # -- plot

        plot = BodePlot(figsize=self.figsize, dpi=self.dpi,
                        dB=self.plot_dB)
        # ax = plot.gca()
        self.tfs = []

        # calculate transfer function
        for seg in groups:
            refts = groups[seg].pop(self.ref_chan)
            for name in groups[seg]:
                series = groups[seg][name]
                self.test_chan = name
                tf = series.transfer_function(refts, fftlength=fftlength,
                                              overlap=overlap,
                                              window=args.window)

                label = name
                if len(self.start_list) > 1:
                    label += f', {series.epoch.gps}'
                if self.usetex:
                    label = label_to_latex(label)

                plot.add_frequencyseries(tf, dB=self.plot_dB, label=label)
                self.tfs.append(tf)

        if args.xscale == 'log' and not args.xmin:
            args.xmin = 1/fftlength

        return plot



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides methods to calculate the astrophysical sensitive
distance of an instrumental power-spectral density.
"""

import warnings

from functools import wraps
from math import pi

from scipy.integrate import trapezoid
from scipy.interpolate import interp1d

from astropy import (
    units,
    constants,
)

from ..spectrogram import Spectrogram
from ..timeseries import TimeSeries
from ..utils import round_to_power

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Alex Urban <alexander.urban@ligo.org>'

DEFAULT_FFT_METHOD = "median"


def _get_isco_frequency(mass1, mass2):
    """Determine the innermost stable circular orbit (ISCO) frequency

    Parameters
    ----------
    mass1 : `float`
        the mass (in solar masses) of the first binary component

    mass2 : `float`
        the mass (in solar masses) of the second binary component

    Returns
    -------
    fisco : `~astropy.units.Quantity`
        linear frequency (Hz) of the innermost stable circular orbit
    """
    mtotal = units.Quantity(mass1 + mass2, 'solMass').to('kg')
    return (
        constants.c ** 3 / (constants.G * 6**1.5 * pi * mtotal)
    ).to('Hz')


def _get_spectrogram(hoft, **kwargs):
    """Check that the input is a spectrogram, or compute one if compatible

    Parameters
    ----------
    hoft : `~gwpy.timeseries.TimeSeries` or `~gwpy.spectrogram.Spectrogram`
        record of gravitational-wave strain output from a detector

    **kwargs : `dict`, optional
        additional keyword arguments to
        `~gwpy.timeseries.TimeSeries.spectrogram`

    Returns
    -------
    hoft : `~gwpy.spectrogram.Spectrogram`
        a time-frequency `Spectrogram` of the input
    """
    if not isinstance(hoft, Spectrogram):
        try:
            hoft = hoft.spectrogram(**kwargs)
        except (AttributeError, TypeError):
            msg = ('Could not produce a spectrogram from the input, please '
                   'pass an instance of gwpy.timeseries.TimeSeries or '
                   'gwpy.spectrogram.Spectrogram')
            raise TypeError(msg)
    return hoft


def _preformat_psd(func):
    @wraps(func)
    def decorated_func(psd, *args, **kwargs):
        if psd.unit != 1/units.Hz:  # force PSD to have the right units
            psd = psd.view()
            psd.override_unit('1/Hz')
        return func(psd, *args, **kwargs)
    return decorated_func


# -- SenseMon range -----------------------------

@_preformat_psd
def sensemon_range_psd(psd, snr=8, mass1=1.4, mass2=1.4, horizon=False):
    """Approximate the inspiral sensitive distance PSD from a GW strain PSD

    This method returns the power spectral density (in ``Mpc**2 / Hz``) to
    which a compact binary inspiral with the given component masses would
    be detectable given the instrumental PSD. The calculation is defined in:

    https://dcc.ligo.org/LIGO-T030276/public

    Parameters
    ----------
    psd : `~gwpy.frequencyseries.FrequencySeries`
        the instrumental power-spectral-density data

    snr : `float`, optional
        the signal-to-noise ratio for which to calculate range,
        default: `8`

    mass1 : `float`, `~astropy.units.Quantity`, optional
        the mass (`float` assumed in solar masses) of the first binary
        component, default: `1.4`

    mass2 : `float`, `~astropy.units.Quantity`, optional
        the mass (`float` assumed in solar masses) of the second binary
        component, default: `1.4`

    horizon : `bool`, optional
        if `True`, return the maximal 'horizon' sensitive distance, otherwise
        return the angle-averaged range, default: `False`

    Returns
    -------
    rspec : `~gwpy.frequencyseries.FrequencySeries`
        the calculated inspiral sensitivity PSD [Mpc^2 / Hz]
    """
    frange = (psd.frequencies > 0)
    # compute total mass and chirp mass
    mass1 = units.Quantity(mass1, 'solMass').to('kg')
    mass2 = units.Quantity(mass2, 'solMass').to('kg')
    mtotal = mass1 + mass2
    mchirp = (mass1 * mass2) ** (3/5.) / mtotal ** (1/5.)
    # calculate integrand with pre-factor
    prefactor = (
        (  # numerator
            (16 if horizon else 1.77**2)
            * 5
            * constants.c ** (1/3.)
            * (mchirp * constants.G / constants.c ** 2) ** (5/3.)
        )  # denominator
        / (
            96
            * pi ** (4/3.)
            * snr ** 2
        )
    )
    return (  # inspiral range PSD, avoiding DC value
        1 / psd[frange] * psd.frequencies[frange] ** (-7/3.) * prefactor
    ).to('Mpc^2 / Hz')


def sensemon_range(psd, snr=8, mass1=1.4, mass2=1.4, fmin=None, fmax=None,
                   horizon=False):
    """Approximate the inspiral sensitive distance from a GW strain PSD

    This method returns the distance (in megaparsecs) to which a compact
    binary inspiral with the given component masses would be detectable
    given the instrumental PSD. The calculation is as defined in:

    https://dcc.ligo.org/LIGO-T030276/public

    Parameters
    ----------
    psd : `~gwpy.frequencyseries.FrequencySeries`
        the instrumental power-spectral-density data

    snr : `float`, optional
        the signal-to-noise ratio for which to calculate range,
        default: `8`

    mass1 : `float`, `~astropy.units.Quantity`, optional
        the mass (`float` assumed in solar masses) of the first binary
        component, default: `1.4`

    mass2 : `float`, `~astropy.units.Quantity`, optional
        the mass (`float` assumed in solar masses) of the second binary
        component, default: `1.4`

    fmin : `float`, optional
        the lower frequency cut-off of the integral, default: `psd.df`

    fmax : `float`, optional
        the maximum frequency limit of the integral, defaults to
        innermost stable circular orbit (ISCO) frequency

    horizon : `bool`, optional
        if `True`, return the maximal 'horizon' sensitive distance, otherwise
        return the angle-averaged range, default: `False`

    Returns
    -------
    range : `~astropy.units.Quantity`
        the calculated inspiral range [Mpc]

    Examples
    --------
    Grab some data for LIGO-Livingston around GW150914 and generate a PSD:

    >>> from gwpy.timeseries import TimeSeries
    >>> hoft = TimeSeries.fetch_open_data('H1', 1126259446, 1126259478)
    >>> hoff = hoft.psd(fftlength=4)

    Now we can calculate the :func:`sensemon_range`:

    >>> from gwpy.astro import sensemon_range
    >>> r = sensemon_range(hoff, fmin=30)
    >>> print(r)
    70.4612102889 Mpc
    """
    fisco = _get_isco_frequency(mass1, mass2)
    # format frequency limits
    fmin = units.Quantity(fmin or psd.df, 'Hz')  # avoid DC value
    fmax = units.Quantity(fmax or fisco, 'Hz')
    if fmax > fisco:
        warnings.warn("Upper frequency bound greater than %s-%s ISCO "
                      "frequency of %s, using ISCO" % (mass1, mass2, fisco))
        fmax = fisco
    # integrate and return
    f = psd.frequencies.to('Hz')
    frange = (f >= fmin) & (f < fmax)
    integrand = sensemon_range_psd(psd[frange], snr=snr, mass1=mass1,
                                   mass2=mass2, horizon=horizon)
    return (units.Quantity(
        trapezoid(integrand.value, f.value[frange]),
        unit=integrand.unit * units.Hertz,
    ) ** (1/2.)).to('Mpc')


# -- inspiral range -----------------------------

def _import_inspiral_range():
    try:
        from inspiral_range import (
            range,
            find_root_redshift,
        )
    except ModuleNotFoundError as exc:  # no inspiral-range package
        exc.msg = str(exc) + (
            "; gwpy.astro's inspiral_range and inspiral_range_psd functions "
            "require the extra package 'inspiral-range' to provide "
            "cosmologically-corrected distance calculations, please install "
            "that package and try again, or install gwpy with the 'astro' "
            "extra via `python -m pip install gwpy[astro]`"
        )
        raise

    from inspiral_range.waveform import CBCWaveform

    return (
        range,
        find_root_redshift,
        CBCWaveform,
    )


@_preformat_psd
def inspiral_range_psd(psd, snr=8, mass1=1.4, mass2=1.4, horizon=False,
                       **kwargs):
    """Calculate the cosmology-corrected inspiral sensitive distance PSD

    This method returns the power spectral density (in ``Mpc**2 / Hz``) to
    which a compact binary inspiral with the given component masses would
    be detectable given the instrumental PSD. The calculation is defined in
    Belczynski et. al (2014):

    https://dx.doi.org/10.1088/0004-637x/789/2/120

    Parameters
    ----------
    psd : `~gwpy.frequencyseries.FrequencySeries`
        the instrumental power-spectral-density data

    snr : `float`, optional
        the signal-to-noise ratio for which to calculate range,
        default: `8`

    mass1 : `float`, `~astropy.units.Quantity`, optional
        the mass (`float` assumed in solar masses) of the first binary
        component, default: `1.4`

    mass2 : `float`, `~astropy.units.Quantity`, optional
        the mass (`float` assumed in solar masses) of the second binary
        component, default: `1.4`

    horizon : `bool`, optional
        if `True`, return the maximal 'horizon' luminosity distance, otherwise
        return the angle-averaged comoving distance, default: `False`

    **kwargs : `dict`, optional
        additional keyword arguments to `~inspiral_range.waveform.CBCWaveform`

    Returns
    -------
    rspec : `~gwpy.frequencyseries.FrequencySeries`
        the calculated inspiral sensitivity PSD [Mpc^2 / Hz]

    See also
    --------
    sensemon_range_psd
        for the method based on LIGO-T030276, also known as LIGO SenseMonitor
    inspiral-range
        the package which does heavy lifting for waveform simulation and
        cosmology calculations
    """
    # import the optional extras required here:
    range_func, find_root_redshift, CBCWaveform = _import_inspiral_range()

    # generate a waveform for this system
    f = psd.frequencies.to('Hz').value
    inspiral = CBCWaveform(f[f > 0], m1=mass1, m2=mass2, **kwargs)

    # determine the detector horizon redshift
    z_hor = find_root_redshift(
        lambda z: inspiral.SNR(psd.value[f > 0], z) - snr,
    )
    if horizon:
        dist = inspiral.cosmo.luminosity_distance(z_hor)
    else:
        dist = range_func(
            f[f > 0],
            psd.value[f > 0],
            z_hor=z_hor,
            H=inspiral,
            detection_snr=snr,
        )

    # calculate the sensitive distance PSD
    (fz, hz) = inspiral.z_scale(z_hor)
    hz = interp1d(fz, hz, bounds_error=False, fill_value=(hz[0], 0))(f)
    out = type(psd)(4 * (dist / snr)**2 * (hz**2 / psd.value)[f > 0])
    # finalize properties and return
    out.__array_finalize__(psd)
    out.override_unit('Mpc^2 / Hz')
    out.f0 = f[f > 0][0]
    return out


@_preformat_psd
def inspiral_range(psd, snr=8, mass1=1.4, mass2=1.4, fmin=None, fmax=None,
                   horizon=False, **kwargs):
    """Calculate the cosmology-corrected inspiral sensitive distance

    This method returns the distance (in megaparsecs) to which a compact
    binary inspiral with the given component masses would be detectable
    given the instrumental PSD. The calculation is defined in Belczynski
    et. al (2014):

    https://dx.doi.org/10.1088/0004-637x/789/2/120

    Parameters
    ----------
    psd : `~gwpy.frequencyseries.FrequencySeries`
        the instrumental power-spectral-density data

    snr : `float`, optional
        the signal-to-noise ratio for which to calculate range,
        default: `8`

    mass1 : `float`, `~astropy.units.Quantity`, optional
        the mass (`float` assumed in solar masses) of the first binary
        component, default: `1.4`

    mass2 : `float`, `~astropy.units.Quantity`, optional
        the mass (`float` assumed in solar masses) of the second binary
        component, default: `1.4`

    fmin : `float`, optional
        the lower frequency cut-off of the integral, default: `psd.df`

    fmax : `float`, optional
        the maximum frequency limit of the integral, defaults to the rest-frame
        innermost stable circular orbit (ISCO) frequency

    horizon : `bool`, optional
        if `True`, return the maximal 'horizon' luminosity distance, otherwise
        return the angle-averaged comoving distance, default: `False`

    **kwargs : `dict`, optional
        additional keyword arguments to `~inspiral_range.waveform.CBCWaveform`

    Returns
    -------
    range : `~astropy.units.Quantity`
        the calculated inspiral range [Mpc]

    Examples
    --------
    Grab some data for LIGO-Livingston around GW150914 and generate a PSD:

    >>> from gwpy.timeseries import TimeSeries
    >>> hoft = TimeSeries.fetch_open_data('H1', 1126259446, 1126259478)
    >>> hoff = hoft.psd(fftlength=4)

    Now, we can calculate the :func:`inspiral_range`:

    >>> from gwpy.astro import inspiral_range
    >>> r = inspiral_range(hoff, fmin=30)
    >>> print(r)
    70.4612102889 Mpc

    See also
    --------
    sensemon_range
        for the method based on LIGO-T030276, also known as LIGO SenseMonitor
    inspiral-range
        the package which does heavy lifting for waveform simulation and
        cosmology calculations
    """
    # import the optional extras required here:
    range_func, find_root_redshift, CBCWaveform = _import_inspiral_range()

    # format frequency limits
    f = psd.frequencies.to('Hz').value
    fmin = fmin or psd.df.value  # avoid DC value
    fmax = fmax or round_to_power(f[-1], which='lower')
    frange = (f >= fmin) & (f < fmax)

    # generate a waveform for this system
    inspiral = CBCWaveform(f[frange], m1=mass1, m2=mass2, **kwargs)

    # determine the detector horizon redshift
    z_hor = find_root_redshift(
        lambda z: inspiral.SNR(psd.value[frange], z) - snr,
    )

    # return the sensitive distance metric
    if horizon:
        dist = inspiral.cosmo.luminosity_distance(z_hor)
    else:
        dist = range_func(
            f[frange],
            psd.value[frange],
            z_hor=z_hor,
            H=inspiral,
            detection_snr=snr,
        )
    return units.Quantity(dist, unit="Mpc")


# -- burst range --------------------------------

@_preformat_psd
def burst_range_spectrum(psd, snr=8, energy=1e-2):
    """Calculate the frequency-dependent GW burst range from a strain PSD

    Parameters
    ----------
    psd : `~gwpy.frequencyseries.FrequencySeries`
        the instrumental power-spectral-density data

    snr : `float`, optional
        the signal-to-noise ratio for which to calculate range,
        default: `8`

    energy : `float`, optional
        the relative energy output of the GW burst,
        default: `0.01` (GRB-like burst)

    Returns
    -------
    rangespec : `~gwpy.frequencyseries.FrequencySeries`
        the burst range `FrequencySeries` [Mpc (default)]
    """
    frange = (psd.frequencies > 0)
    # calculate frequency dependent range in parsecs
    a = (
        constants.G * energy * constants.M_sun * 0.4
        / (pi**2 * constants.c)
    ) ** (1/2.)
    return (  # burst range spectrum, avoiding DC value
        psd[frange] ** (-1/2.) * a / (snr * psd.frequencies[frange])
    ).to('Mpc')


def burst_range(psd, snr=8, energy=1e-2, fmin=100, fmax=500):
    """Calculate the integrated GRB-like GW burst range from a strain PSD

    Parameters
    ----------
    psd : `~gwpy.frequencyseries.FrequencySeries`
        the instrumental power-spectral-density data

    snr : `float`, optional
        the signal-to-noise ratio for which to calculate range,
        default: ``8``

    energy : `float`, optional
        the relative energy output of the GW burst, defaults to ``1e-2``
        for a GRB-like burst

    fmin : `float`, optional
        the lower frequency cutoff of the burst range integral,
        default: ``100 Hz``

    fmax : `float`, optional
        the upper frequency cutoff of the burst range integral,
        default: ``500 Hz``

    Returns
    -------
    range : `~astropy.units.Quantity`
        the GRB-like-burst sensitive range [Mpc (default)]

    Examples
    --------
    Grab some data for LIGO-Livingston around GW150914 and generate a PSD

    >>> from gwpy.timeseries import TimeSeries
    >>> hoft = TimeSeries.fetch_open_data('H1', 1126259446, 1126259478)
    >>> hoff = hoft.psd(fftlength=4)

    Now we can calculate the :func:`burst_range`:

    >>> from gwpy.astro import burst_range
    >>> r = burst_range(hoff, fmin=30)
    >>> print(r)
    42.5055584195 Mpc
    """
    f = psd.frequencies.to('Hz').value
    # restrict integral
    fmin = fmin or psd.df.to('Hz').value
    fmax = fmax or f[-1].value
    frange = (f >= fmin) & (f < fmax)
    # calculate integrand and integrate
    integrand = burst_range_spectrum(
        psd[frange], snr=snr, energy=energy) ** 3
    out = trapezoid(integrand.value, f[frange])
    # normalize and return
    return (units.Quantity(
        out / (fmax - fmin),
        unit=integrand.unit,
    ) ** (1/3.)).to('Mpc')


# -- timeseries/spectrogram wrappers ------------

def range_timeseries(
    hoft,
    stride=None,
    fftlength=None,
    overlap=None,
    window='hann',
    method=DEFAULT_FFT_METHOD,
    nproc=1,
    range_func=None,
    **rangekwargs,
):
    """Measure timeseries trends of astrophysical detector range (Mpc)
    directly from strain

    Parameters
    ----------
    hoft : `~gwpy.timeseries.TimeSeries` or `~gwpy.spectrogram.Spectrogram`
        record of gravitational-wave strain output from a detector

    stride : `float`, optional
        desired step size (seconds) of range timeseries, required if
        `hoft` is an instance of `TimeSeries`

    fftlength : `float`, optional
        number of seconds in a single FFT

    overlap : `float`, optional
        number of seconds of overlap between FFTs, defaults to the
        recommended overlap for the given window (if given), or 0

    window : `str`, `numpy.ndarray`, optional
        window function to apply to timeseries prior to FFT, see
        :func:`scipy.signal.get_window` for details on acceptable
        formats

    method : `str`, optional
        FFT-averaging method, defaults to median averaging, see
        :meth:`~gwpy.timeseries.TimeSeries.spectrogram` for
        more details

    nproc : `int`, optional
        number of CPUs to use in parallel processing of FFTs, default: 1

    range_func : `callable`, optional
        the function to call to generate the range for each stride,
        defaults to ``inspiral_range`` unless ``energy`` is given
        as a keyword argument to the range function

    **rangekwargs : `dict`, optional
        additional keyword arguments to :func:`burst_range` or
        :func:`inspiral_range` (see "Notes" below), defaults to
        inspiral range with `mass1 = mass2 = 1.4` solar masses

    Returns
    -------
    out : `~gwpy.timeseries.TimeSeries`
        timeseries trends of astrophysical range

    Notes
    -----
    This method is designed to quantify a gravitational-wave detector's
    sensitive range as a function of time. It supports the range to
    compact binary inspirals and to unmodelled GW bursts, each a class
    of transient event.

    See also
    --------
    gwpy.timeseries.TimeSeries.spectrogram
        for the underlying power spectral density estimator
    inspiral_range
        for the function that computes inspiral range
    burst_range
        for the function that computes burst range
    range_spectrogram
        for a `~gwpy.spectrogram.Spectrogram` of the range integrand
    """
    rangekwargs = rangekwargs or {'mass1': 1.4, 'mass2': 1.4}
    if range_func is None and "energy" in rangekwargs:
        range_func = burst_range
    elif range_func is None:
        range_func = inspiral_range
    hoft = _get_spectrogram(
        hoft, stride=stride, fftlength=fftlength, overlap=overlap,
        window=window, method=method, nproc=nproc)
    # loop over time bins
    out = TimeSeries(
        [range_func(psd, **rangekwargs).value for psd in hoft],
    )
    # finalise output
    out.__array_finalize__(hoft)
    out.override_unit('Mpc')
    return out


def range_spectrogram(
    hoft,
    stride=None,
    fftlength=None,
    overlap=None,
    window='hann',
    method=DEFAULT_FFT_METHOD,
    nproc=1,
    range_func=None,
    **rangekwargs,
):
    """Calculate the average range or range power spectrogram (Mpc or
    Mpc^2 / Hz) directly from strain

    Parameters
    ----------
    hoft : `~gwpy.timeseries.TimeSeries`  or `~gwpy.spectrogram.Spectrogram`
        record of gravitational-wave strain output from a detector

    stride : `float`, optional
        number of seconds in a single PSD (i.e., step size of spectrogram),
        required if `hoft` is an instance of `TimeSeries`

    fftlength : `float`, optional
        number of seconds in a single FFT

    overlap : `float`, optional
        number of seconds of overlap between FFTs, defaults to the
        recommended overlap for the given window (if given), or 0

    window : `str`, `numpy.ndarray`, optional
        window function to apply to timeseries prior to FFT, see
        :func:`scipy.signal.get_window` for details on acceptable
        formats

    method : `str`, optional
        FFT-averaging method, defaults to median averaging, see
        :meth:`~gwpy.timeseries.TimeSeries.spectrogram` for
        more details

    nproc : `int`, optional
        number of CPUs to use in parallel processing of FFTs, default: 1

    fmin : `float`, optional
        low frequency cut-off (Hz), defaults to `1/fftlength`

    fmax : `float`, optional
        high frequency cut-off (Hz), defaults to Nyquist frequency of `hoft`

    range_func : `callable`, optional
        the function to call to generate the range for each stride,
        defaults to ``inspiral_range`` unless ``energy`` is given
        as a keyword argument to the range function

    **rangekwargs : `dict`, optional
        additional keyword arguments to :func:`burst_range_spectrum` or
        :func:`inspiral_range_psd` (see "Notes" below), defaults to
        inspiral range with `mass1 = mass2 = 1.4` solar masses

    Returns
    -------
    out : `~gwpy.spectrogram.Spectrogram`
        time-frequency spectrogram of astrophysical range

    Notes
    -----
    This method is designed to show the contribution to a
    gravitational-wave detector's sensitive range across frequency bins
    as a function of time. It supports the range to compact binary
    inspirals and to unmodelled GW bursts, each a class of transient
    event.

    If inspiral range is requested and `fmax` exceeds the frequency of the
    innermost stable circular orbit (ISCO), the output will extend only up
    to the latter.

    See also
    --------
    gwpy.timeseries.TimeSeries.spectrogram
        for the underlying power spectral density estimator
    inspiral_range_psd
        for the function that computes inspiral range integrand
    burst_range_spectrum
        for the function that computes burst range integrand
    range_timeseries
        for `TimeSeries` trends of the astrophysical range
    """
    rangekwargs = rangekwargs or {'mass1': 1.4, 'mass2': 1.4}
    if range_func is None and "energy" in rangekwargs:
        range_func = burst_range_spectrum
    elif range_func is None:
        range_func = inspiral_range_psd
    hoft = _get_spectrogram(
        hoft, stride=stride, fftlength=fftlength, overlap=overlap,
        window=window, method=method, nproc=nproc)
    # set frequency limits
    f = hoft.frequencies.to('Hz')
    fmin = units.Quantity(
        rangekwargs.pop('fmin', hoft.df),
        'Hz',
    )
    fmax = units.Quantity(
        rangekwargs.pop(
            'fmax',
            round_to_power(f[-1].value, which='lower'),
        ),
        'Hz',
    )
    frange = (f >= fmin) & (f < fmax)
    # loop over time bins
    out = Spectrogram(
        [range_func(psd[frange], **rangekwargs).value for psd in hoft],
    )
    # finalise output
    out.__array_finalize__(hoft)
    out.override_unit('Mpc' if 'energy' in rangekwargs
                      else 'Mpc^2 / Hz')
    out.f0 = fmin
    return out



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for utils module
"""

import pytest

import numpy

from astropy import units

# import necessary modules
lal = pytest.importorskip('lal')
utils_lal = pytest.importorskip('gwpy.utils.lal')

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def test_to_lal_type_str():
    assert utils_lal.to_lal_type_str(float) == 'REAL8'
    assert utils_lal.to_lal_type_str(
        numpy.dtype('float64')) == 'REAL8'
    assert utils_lal.to_lal_type_str(11) == 'REAL8'
    with pytest.raises(ValueError):
        utils_lal.to_lal_type_str('blah')
    with pytest.raises(ValueError):
        utils_lal.to_lal_type_str(numpy.int8)
    with pytest.raises(ValueError):
        utils_lal.to_lal_type_str(20)


def test_find_typed_function():
    assert utils_lal.find_typed_function(
        'REAL8', 'Create', 'Sequence') is lal.CreateREAL8Sequence

    try:
        import lalframe
    except ImportError:  # no lalframe
        return
    utils_lal.find_typed_function(
        'REAL4', 'FrStreamRead', 'TimeSeries',
        module=lalframe) is lalframe.FrStreamReadREAL4TimeSeries


@pytest.mark.parametrize(
    'dtype',
    [
        numpy.int16,
        numpy.int32,
        numpy.int64,
        numpy.uint16,
        numpy.uint32,
        numpy.uint64,
        numpy.float32,
        numpy.float64,
        numpy.complex64,
        numpy.complex128,
    ]
)
def test_from_lal_type(dtype):
    func = utils_lal.find_typed_function(dtype, 'Create', 'TimeSeries')
    lalts = func(None, None, 0, 1., None, 1)
    assert utils_lal.from_lal_type(lalts) is dtype
    assert utils_lal.from_lal_type(type(lalts)) is dtype


@pytest.mark.parametrize('name', [
    'XXX',
    'XXXCOMPLEX64ZZZ',
    'INNT2',
    'INT5',
    'REAL42',
])
def test_from_lal_type_errors(name):
    lal_type = type(name, (), {})
    with pytest.raises(ValueError, match='no known numpy'):
        utils_lal.from_lal_type(lal_type)


@pytest.mark.parametrize(("in_", "out", "scale"), (
    ("m", "m", 1),
    ("Farad", "m^-2 kg^-1 s^4 A^2", 1),
    ("m**(1/2)", "m^1/2", 1),
    ("km", "10^3 m", 1),
    ("123 m", "m", 123),
))
def test_to_lal_unit(in_, out, scale):
    assert utils_lal.to_lal_unit(in_) == (lal.Unit(out), scale)


def test_to_lal_unit_error():
    with pytest.raises(
        ValueError,
        match="^LAL has no unit corresponding to 'rad'$",
    ):
        utils_lal.to_lal_unit('rad/s')


@pytest.mark.parametrize(("in_", "out"), (
    ("m s^-1", "m/s"),
    ("strain", "strain"),
    ("m^1/2", "m**(1/2.)"),
    ("10^3 m", "km"),
))
def test_from_lal_unit(in_, out):
    assert utils_lal.from_lal_unit(lal.Unit(in_)) == units.Unit(out)


def test_to_lal_ligotimegps():
    assert utils_lal.to_lal_ligotimegps(123.456) == (
        lal.LIGOTimeGPS(123, 456000000))



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.cli.spectrum`
"""

from astropy.time import Time

from ... import cli
from .base import _TestFrequencyDomainProduct


class TestCliSpectrum(_TestFrequencyDomainProduct):
    TEST_CLASS = cli.Spectrum
    ACTION = 'spectrum'

    def test_get_title(self, prod):
        epoch = prod.start_list[0]
        utc = Time(epoch, format='gps', scale='utc').iso
        t = ', '.join([
            f'{utc} | {epoch} ({prod.duration})',
            f'fftlength={prod.args.secpfft}',
            f'overlap={prod.args.overlap}',
        ])
        assert prod.get_title() == t

    def test_get_suptitle(self, prod):
        assert prod.get_suptitle() == f'Spectrum: {prod.chan_list[0]}'



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.timeseries.io.gwf.lalframe`
"""

from pathlib import Path
from urllib.parse import urlparse

import pytest
import numpy

from gwdatafind.utils import file_segment

from ...io.cache import write_cache
from ...segments import Segment
from ...testing.utils import (
    assert_dict_equal,
    assert_quantity_sub_equal,
    TEST_GWF_FILE,
)
from ...timeseries import TimeSeries

# import optional dependencies
lal_utils = pytest.importorskip("lal.utils")
lalframe = pytest.importorskip("lalframe")
gwpy_lalframe = pytest.importorskip("gwpy.timeseries.io.gwf.lalframe")

# get URI to test against
TEST_GWF_PATH = Path(TEST_GWF_FILE).absolute()
TEST_GWF_URL = TEST_GWF_PATH.as_uri()

# get info corresponding to this file
TEST_GWF_SEGMENT = file_segment(TEST_GWF_FILE)
TEST_GWF_DELTA_T = 1 / 16384

# channels to read
CHANNELS = ["H1:LDAS-STRAIN", "L1:LDAS-STRAIN", "V1:h_16384Hz"]


@pytest.fixture
def stream():
    return lalframe.FrStreamOpen(str(TEST_GWF_PATH.parent), TEST_GWF_PATH.name)


def _test_open_data_source(source):
    """This function actually performs the test
    """
    stream = gwpy_lalframe.open_data_source(source)
    assert stream.epoch == TEST_GWF_SEGMENT[0]
    assert Path(urlparse(stream.cache.list.url).path).samefile(TEST_GWF_PATH)


@pytest.mark.parametrize("source", [
    TEST_GWF_FILE,
    TEST_GWF_URL,
    [TEST_GWF_FILE],
    lal_utils.CacheEntry.from_T050017(TEST_GWF_FILE),
])
def test_open_data_source(source):
    return _test_open_data_source(source)


@pytest.mark.requires("glue.lal")
def test_open_data_source_glue():
    from glue.lal import Cache
    Cache.entry_class = lal_utils.CacheEntry
    cache = Cache.from_urls([TEST_GWF_FILE])
    return _test_open_data_source(cache)


def test_open_data_source_cache(tmp_path):
    tmp = tmp_path / "test.lcf"
    write_cache([TEST_GWF_FILE], tmp, format="lal")
    return _test_open_data_source(tmp)


def test_open_data_source_error():
    with pytest.raises(
        ValueError,
        match=f"^Don't know how to open data source of type {type(None)}$",
    ):
        gwpy_lalframe.open_data_source(None)


def test_get_stream_duration(stream):
    assert gwpy_lalframe.get_stream_duration(stream) == 1.


@pytest.mark.parametrize("start, end", [
    (None, None),
    (None, TEST_GWF_SEGMENT[0] + .5),
    (TEST_GWF_SEGMENT[0] + .5, None),
    (TEST_GWF_SEGMENT[0] + .25, TEST_GWF_SEGMENT[1] - .25),
    (TEST_GWF_SEGMENT[1] - TEST_GWF_DELTA_T / 2, None),
])
def test_read(start, end):
    data = gwpy_lalframe.read(
        TEST_GWF_FILE,
        CHANNELS,
        start=start,
        end=end,
    )
    assert isinstance(data, dict)

    start = TEST_GWF_SEGMENT[0] if start is None else start
    end = TEST_GWF_SEGMENT[1] if end is None else end

    for name in CHANNELS:
        ts = data[name]
        # check basic parameters
        assert ts.sample_rate.value == 16384
        assert ts.name == name

        # check data span is what we asked for
        assert numpy.allclose(ts.xspan, (start, end), atol=TEST_GWF_DELTA_T)


def test_read_channel_error():
    with pytest.raises(
        RuntimeError,
        match="^channel 'bad' not found$",
    ):
        gwpy_lalframe.read(TEST_GWF_FILE, ["bad"])


def test_read_deprecated_scaled():
    with pytest.warns(UserWarning):
        gwpy_lalframe.read(
            TEST_GWF_FILE,
            ["L1:LDAS-STRAIN"],
            scaled=True,
        )


def test_write(tmp_path):
    # read the data first
    data = gwpy_lalframe.read(TEST_GWF_FILE, CHANNELS)

    # write the data
    tmp = tmp_path / "test.gwf"
    gwpy_lalframe.write(
        data,
        tmp,
    )

    # read it back and check things
    data2 = gwpy_lalframe.read(tmp, CHANNELS)
    assert_dict_equal(data, data2, assert_quantity_sub_equal)


@pytest.mark.parametrize(
    'data',
    [
        pytest.param(
            [],
            marks=pytest.mark.xfail(
                raises=RuntimeError,
                reason='Cannot add an empty series to a frame',
            )
        ),
        [1, 2, 3, 4, 5],
    ]
)
def test_write_no_ifo(tmp_path, data):
    # create timeseries with no IFO
    data = TimeSeries(data, dtype=float)
    tmp = tmp_path / "test.gwf"
    gwpy_lalframe.write(
        {None: data},
        tmp
    )


def test_read_missing_sample():
    """Check that giving non-sampled start time doesn't result in missing data.

    See https://git.ligo.org/computing/helpdesk/-/issues/4774 and
    https://git.ligo.org/lscsoft/lalsuite/-/issues/710.
    """
    data = TimeSeries.read(
        TEST_GWF_PATH,
        "H1:LDAS-STRAIN",
        start=968654552.6709,
        end=968654553,
        format="gwf.lalframe",
    )
    assert data.span == Segment(968654552.670898432, 968654553.0)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2018-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwpy.plot.tex`
"""

from unittest import mock

import pytest

from .. import tex as plot_tex


def _which(arg):
    """Fake which to force pdflatex to being not found
    """
    if arg == "pdflatex":
        return None
    return arg


@mock.patch("gwpy.plot.tex.which", _which)
def test_has_tex_missing_exe():
    """Test that `gwpy.plot.tex.has_tex` returns `False` when
    any one of the necessary executables is missing.
    """
    plot_tex.has_tex.cache_clear()
    assert not plot_tex.has_tex()


@mock.patch("gwpy.plot.tex._test_usetex", side_effect=RuntimeError)
def test_has_tex_bad_latex(_):
    """Test that `gwpy.plot.tex.has_tex` returns `False` when
    the LaTeX figure fails to render.
    """
    plot_tex.has_tex.cache_clear()
    assert not plot_tex.has_tex()


@mock.patch("gwpy.plot.tex.which", return_value="path")
@mock.patch("gwpy.plot.tex._test_usetex")
def test_has_tex_true(_which_, _test_usetex):
    """Test that `gwpy.plot.tex.has_tex` returns `True` when
    all of the necessary executables are found, and the LaTeX figure
    doesn't raise an exception.
    """
    plot_tex.has_tex.cache_clear()
    assert plot_tex.has_tex()


@pytest.mark.parametrize('in_, out', [
    (1, '1'),
    (100, r'10^{2}'),
    (-500, r'-5\!\!\times\!\!10^{2}'),
    (0.00001, r'10^{-5}'),
])
def test_float_to_latex(in_, out):
    assert plot_tex.float_to_latex(in_) == out


@pytest.mark.parametrize('in_, out', [
    (None, ''),
    ('normal text', 'normal text'),
    (r'$1 + 2 = 3$', r'$1 + 2 = 3$'),
    ('H1:ABC-DEF_GHI', r'H1:ABC-DEF\_GHI'),
    (r'H1:ABC-DEF\_GHI', r'H1:ABC-DEF\_GHI'),
])
def test_label_to_latex(in_, out):
    assert plot_tex.label_to_latex(in_) == out



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for `gwpy.io.ligolw` module
"""

import tempfile
from pathlib import Path

import pytest

import numpy

from .. import ligolw as io_ligolw

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

OLD_FORMAT_LIGO_LW_XML = """
<?xml version='1.0' encoding='utf-8'?>
<!DOCTYPE LIGO_LW SYSTEM "http://ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt">
<LIGO_LW>
  <Table Name="sngl_burst:table">
    <Column Type="lstring" Name="sngl_burst:ifo"/>
    <Column Type="int_4s" Name="sngl_burst:peak_time"/>
    <Column Type="int_4s" Name="sngl_burst:peak_time_ns"/>
    <Column Type="int_4s" Name="sngl_burst:start_time"/>
    <Column Type="int_4s" Name="sngl_burst:start_time_ns"/>
    <Column Type="real_4" Name="sngl_burst:duration"/>
    <Column Type="lstring" Name="sngl_burst:search"/>
    <Column Type="ilwd:char" Name="sngl_burst:event_id"/>
    <Column Type="ilwd:char" Name="sngl_burst:process_id"/>
    <Column Type="real_4" Name="sngl_burst:central_freq"/>
    <Column Type="lstring" Name="sngl_burst:channel"/>
    <Column Type="real_4" Name="sngl_burst:amplitude"/>
    <Column Type="real_4" Name="sngl_burst:snr"/>
    <Column Type="real_4" Name="sngl_burst:confidence"/>
    <Column Type="real_8" Name="sngl_burst:chisq"/>
    <Column Type="real_8" Name="sngl_burst:chisq_dof"/>
    <Column Type="real_4" Name="sngl_burst:bandwidth"/>
    <Stream Delimiter="," Type="Local" Name="sngl_burst:table">
      "H1",100000000,000000000,100000000,000000000,1,"gstlal_excesspower","sngl_burst:event_id:4696","process:process_id:0",162,"LDAS-STRAIN",2.3497068e-25,0.69409615,16.811825,0,512,256,
      "H1",100000000,000000001,100000000,000000001,1,"gstlal_excesspower","sngl_burst:event_id:4697","process:process_id:0",162,"LDAS-STRAIN",2.3497797e-25,0.69416249,16.816761,0,512,256,
      "H1",100000000,000000001,100000000,000000002,1,"gstlal_excesspower","sngl_burst:event_id:4698","process:process_id:0",162,"LDAS-STRAIN",2.3479907e-25,0.69253588,16.696106,0,512,256,
    </Stream>
  </Table>
</LIGO_LW>
""".strip()  # noqa: E501


# -- fixtures -----------------------------------------------------------------

@pytest.fixture(scope="function")
def llwdoc():
    """Build an empty LIGO_LW Document
    """
    try:
        from ligo.lw.ligolw import (Document, LIGO_LW)
    except ImportError as exc:
        pytest.skip(str(exc))
    xmldoc = Document()
    xmldoc.appendChild(LIGO_LW())
    return xmldoc


def new_table(tablename, data=None, **new_kw):
    """Create a new LIGO_LW Table with data
    """
    try:
        from ligo.lw import lsctables
    except ImportError as exc:
        pytest.skip(str(exc))
    from ligo.lw.table import Table

    table = lsctables.New(
        lsctables.TableByName[Table.TableName(tablename)],
        **new_kw,
    )
    for dat in data or list():
        row = table.RowType()
        for key, val in dat.items():
            setattr(row, key, val)
        table.append(row)
    return table


@pytest.fixture(scope="function")
def llwdoc_with_tables(llwdoc):
    """Build a LIGO_LW Document with some tables
    """
    llw = llwdoc.childNodes[-1]  # get ligolw element
    for t in [
        new_table('process'),
        new_table('sngl_ringdown'),
    ]:
        llw.appendChild(t)
    return llwdoc


# -- tests --------------------------------------------------------------------

def test_read_table(llwdoc_with_tables):
    tab = io_ligolw.read_table(llwdoc_with_tables, tablename="process")
    assert tab is llwdoc_with_tables.childNodes[0].childNodes[0]


def test_read_table_empty(llwdoc):
    with pytest.raises(
        ValueError,
        match=r"^No tables found in LIGO_LW document\(s\)$",
    ):
        io_ligolw.read_table(llwdoc)


@pytest.mark.requires("ligo.lw.lsctables")
def test_read_table_ilwd(tmp_path):
    xmlpath = tmp_path / "test.xml"
    with open(xmlpath, "w") as f:
        f.write(OLD_FORMAT_LIGO_LW_XML)
    tab = io_ligolw.read_table(xmlpath, tablename="sngl_burst")
    assert len(tab) == 3


def test_read_table_multiple(llwdoc_with_tables):
    """Check that `gwpy.io.ligolw.read_table` correctly errors on ambiguity.
    """
    with pytest.raises(
        ValueError,
        match="^Multiple tables .* 'process', 'sngl_ringdown'",
    ):
        io_ligolw.read_table(llwdoc_with_tables)


def test_open_xmldoc(tmp_path, llwdoc_with_tables):
    tmp = tmp_path / "test.xml"
    # write a LIGO_LW file
    with open(tmp, "w") as fobj:
        llwdoc_with_tables.write(fobj)
    # and check that we can read it again (from Path, str, and file)
    for obj in (tmp, str(tmp), open(tmp, "rb")):
        copy = io_ligolw.open_xmldoc(obj)
        # and that the contents are the same
        assert (
            copy.childNodes[0].childNodes
            == llwdoc_with_tables.childNodes[0].childNodes
        )


def test_open_xmldoc_new(tmp_path, llwdoc):
    from ligo.lw.ligolw import Document
    new = io_ligolw.open_xmldoc(tmp_path / "new.xml")
    assert isinstance(new, Document)
    assert not new.childNodes  # empty


def test_get_ligolw_element(llwdoc):
    llw = llwdoc.childNodes[0]
    assert io_ligolw.get_ligolw_element(llw) is llw
    assert io_ligolw.get_ligolw_element(llwdoc) is llw


def test_get_ligolw_element_error(llwdoc):
    # check that blank document raises an error
    with pytest.raises(ValueError):
        io_ligolw.get_ligolw_element(type(llwdoc)())


def test_iter_tables(llwdoc_with_tables):
    expected = llwdoc_with_tables.childNodes[0].childNodes
    assert list(io_ligolw.iter_tables(llwdoc_with_tables)) == expected


def test_list_tables(llwdoc_with_tables):
    names = [
        t.TableName(t.Name)
        for t in llwdoc_with_tables.childNodes[0].childNodes
    ]

    # check that tables are listed properly
    assert io_ligolw.list_tables(llwdoc_with_tables) == names


def test_list_tables_file(llwdoc_with_tables):
    # check that we can list from files
    names = [
        t.TableName(t.Name)
        for t in llwdoc_with_tables.childNodes[0].childNodes
    ]
    with tempfile.NamedTemporaryFile(mode='w') as f:
        llwdoc_with_tables.write(f)
        f.seek(0)
        assert io_ligolw.list_tables(f) == names


@pytest.mark.requires("ligo.lw.lsctables")
@pytest.mark.parametrize('value, name, result', [
    (None, 'peak_time', None),
    (1.0, 'peak_time', numpy.int32(1)),
    (1, 'process_id', 1),
    (1.0, 'invalidname', 1.0),
])
def test_to_table_type(value, name, result):
    from ligo.lw.lsctables import SnglBurstTable
    out = io_ligolw.to_table_type(value, SnglBurstTable, name)
    assert isinstance(out, type(result))
    assert out == result


def test_write_tables_to_document(llwdoc_with_tables):
    # create new table
    def _new():
        return new_table(
            'segment',
            [{'segment': (1, 2)}, {'segment': (3, 4)}, {'segment': (5, 6)}],
            columns=('start_time', 'start_time_ns', 'end_time', 'end_time_ns'),
        )

    # get ligolw element
    llw = llwdoc_with_tables.childNodes[-1]

    # check we can add a new table
    tab = _new()
    io_ligolw.write_tables_to_document(llwdoc_with_tables, [tab])
    assert llw.childNodes[-1] is tab
    assert len(tab) == 3

    # check that adding to an existing table extends
    io_ligolw.write_tables_to_document(llwdoc_with_tables, [_new()])
    assert len(llw.childNodes[-1]) == 6

    # check overwrite=True gives a fresh table
    io_ligolw.write_tables_to_document(
        llwdoc_with_tables,
        [_new()],
        overwrite=True,
    )
    assert len(llw.childNodes[-1]) == 3


def test_write_tables(tmp_path):
    stab = new_table(
        'segment',
        [{'segment': (1, 2)}, {'segment': (3, 4)}, {'segment': (5, 6)}],
        columns=('start_time', 'start_time_ns', 'end_time', 'end_time_ns'),
    )
    ptab = new_table(
        'process',
        [{'program': 'gwpy'}],
        columns=('program',),
    )

    tmp = tmp_path / "test.xml"

    # write writing works
    io_ligolw.write_tables(tmp, [stab, ptab])
    assert io_ligolw.list_tables(tmp) == ['segment', 'process']

    # check writing to existing file raises IOError
    with pytest.raises(IOError):
        io_ligolw.write_tables(tmp, [stab, ptab])

    # check overwrite=True works
    io_ligolw.write_tables(tmp, [stab], overwrite=True)
    xmldoc = io_ligolw.open_xmldoc(tmp)
    assert io_ligolw.list_tables(xmldoc) == ['segment']
    stab2 = io_ligolw.read_table(xmldoc, 'segment')
    assert len(stab2) == len(stab)

    io_ligolw.write_tables(tmp, [stab, ptab], overwrite=True)  # rewrite

    # check append=True works
    io_ligolw.write_tables(tmp, [stab], append=True)
    xmldoc = io_ligolw.open_xmldoc(tmp)
    assert sorted(io_ligolw.list_tables(xmldoc)) == ['process', 'segment']
    stab2 = io_ligolw.read_table(xmldoc, 'segment')
    assert len(stab2) == len(stab) * 2

    # check append=True, overwrite=True works
    io_ligolw.write_tables(tmp, [stab], append=True, overwrite=True)
    xmldoc = io_ligolw.open_xmldoc(tmp)
    assert sorted(io_ligolw.list_tables(xmldoc)) == ['process', 'segment']
    stab2 = io_ligolw.read_table(xmldoc, 'segment')
    assert len(stab2) == len(stab)


@pytest.mark.requires("ligo.lw.lsctables")
def test_is_ligolw_false():
    assert not io_ligolw.is_ligolw("read", None, None, 1)


def test_is_ligolw_obj(llwdoc):
    assert io_ligolw.is_ligolw("read", None, None, llwdoc)


def test_is_ligolw_file(llwdoc):
    with tempfile.TemporaryDirectory() as tmpdir:
        f = str(Path(tmpdir) / "test.xml")
        with open(f, "w"):  # create an empty file
            pass
        # assert that it isn't identified as LIGO_LW XML
        with open(f, "r") as fobj:
            assert not io_ligolw.is_ligolw("read", f, fobj)

        # now write a LIGO_LW file
        with open(f, "w") as fobj:
            llwdoc.write(fobj)
        # and check that it is identified as LIGO_LW XML
        with open(f, "r") as fobj:
            assert io_ligolw.is_ligolw("read", f, fobj)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from ROOT trees into Tables
"""

from ...io import registry
from ...io.utils import (file_path, identify_factory)
from .. import (Table, EventTable)
from .utils import (read_with_columns, read_with_selection)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def _get_treename(source, trees):
    """Return the name of the only tree in the trees

    Raises
    ------
    ValueError
        if multiple trees are found
    """
    if not trees:  # nothing to read?
        raise ValueError("No trees found in %s" % source)

    # get list of all names
    names = [
        name.decode("utf-8") if isinstance(name, bytes) else name
        for name in trees
    ]
    if len(names) == 1:
        return names[0]
    raise ValueError(
        "Multiple trees found in {}, please select one via the "
        "`treename` keyword argument, e.g. `treename='events'`. "
        "Available trees are: '{}'.".format(
            source,
            "', '".join(names),
        ),
    )


@read_with_columns
@read_with_selection
def table_from_root(source, treename=None, **kwargs):
    """Read a Table from a ROOT tree
    """
    import uproot

    path = file_path(source)
    with uproot.open(path) as trees:
        # find tree name
        if treename is None:
            treename = _get_treename(path, trees)

        # read branches from tree
        try:
            return Table(trees[treename].arrays(library="np"), **kwargs)
        except TypeError:  # uproot < 4
            return Table(trees[treename].arrays(namedecode="utf-8"), **kwargs)


def table_to_root(
        table,
        filename,
        treename="tree",
        overwrite=False,
        append=False,
        **kwargs,
):
    """Write a Table to a ROOT file
    """
    import uproot

    createkw = {
        k: kwargs.pop(k) for k in {
            "initial_directory_bytes",
            "uuid_function",
        } if k in kwargs
    }
    if append:
        create_func = uproot.update
    elif overwrite:
        create_func = uproot.recreate
    else:
        create_func = uproot.create

    with create_func(filename, **createkw) as outf:
        tree = outf.mktree(
            treename,
            dict(table.dtype.descr),
            **kwargs,
        )
        tree.extend(dict(table.columns))


# register I/O
for table_class in (Table, EventTable):
    registry.register_reader('root', table_class, table_from_root)
    registry.register_writer('root', table_class, table_to_root)
    registry.register_identifier('root', table_class,
                                 identify_factory('.root'))



# -*- coding: utf-8 -*-
# Copyright (C) Louisiana State University (2014-2017)
#               Cardiff University (2017-2021)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for `gwpy.table.gravityspy`
"""

import pytest

import requests

from ...testing.errors import pytest_skip_network_error
from .. import (
    gravityspy as table_gravityspy,
    GravitySpyTable,
)
from .test_table import TestEventTable as _TestEventTable

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

TEST_ID = "8FHTgA8MEu"
TEST_IFO = "H1"


class TestGravitySpyTable(_TestEventTable):
    TABLE = GravitySpyTable

    @classmethod
    def search(cls):
        return cls.TABLE.search(
            gravityspy_id=TEST_ID,
            howmany=1,
            ifos="H1L1",
            timeout=60,
        )

    @pytest_skip_network_error
    def test_search(self):
        """Test `GravitySpyTable.search`
        """
        # run the search
        table = self.search()

        # validate the result
        assert len(table) == 1
        t = table[0]
        assert t["gravityspy_id"] == TEST_ID
        assert t["ifo"] == TEST_IFO
        assert t["ml_label"] == "Scratchy"

    def test_search_error(self):
        """Test `GravitySpyTable.search` error handling

        Mainly to make sure that an HTTP Error is raised instead
        of a JSONDecodeError.
        """
        requests_mock = pytest.importorskip("requests_mock")
        mocker = requests_mock.Mocker()
        raises = pytest.raises(
            requests.HTTPError,
            match="please check the request parameters$",
        )
        with mocker as rmock, raises:
            rmock.get(
                "{}{}{}/?{}".format(
                    table_gravityspy.DEFAULT_HOST,
                    table_gravityspy.SEARCH_PATH,
                    table_gravityspy.SIMILARITY_SEARCH_PATH,
                    "&".join((
                        "howmany=10",
                        "imageid=abcde",
                        "era=event_time+BETWEEN+1126400000+AND+1584057618",
                        "ifo=%27H1%27%2C+%27L1%27",
                        "database=similarity_index_o3",
                    )),
                ),
                text="<!DOCTYPE html><html></html>",
                status_code=200,
                headers={
                    "Content-Type": "text/html; charset=utf-8",
                },
            )
            self.TABLE.search(gravityspy_id="abcde")

    @pytest_skip_network_error
    def test_download(self, tmp_path):
        """Test `GravitySpyTable.download`
        """
        table = self.search()
        table.download(download_path=tmp_path)
        assert (
            tmp_path
            / "{}_{}_spectrogram_0.5.png".format(TEST_IFO, TEST_ID)
        ).is_file()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017-2020)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Mock objects for GWpy tests
"""

import inspect
from unittest import mock

from ..detector import Channel
from ..time import LIGOTimeGPS

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


# -- NDS2 ---------------------------------------------------------------------

def nds2_buffer(channel, data, epoch, sample_rate, unit,
                name=None, slope=1, offset=0):
    import nds2
    epoch = LIGOTimeGPS(epoch)
    ndsbuffer = mock.create_autospec(nds2.buffer)
    ndsbuffer.length = len(data)
    ndsbuffer.channel = nds2_channel(channel, sample_rate, unit)
    ndsbuffer.name = name or ndsbuffer.channel.name
    ndsbuffer.sample_rate = sample_rate
    ndsbuffer.gps_seconds = epoch.gpsSeconds
    ndsbuffer.gps_nanoseconds = epoch.gpsNanoSeconds
    ndsbuffer.signal_slope = slope
    ndsbuffer.signal_offset = offset
    ndsbuffer.data = data
    return ndsbuffer


def nds2_buffer_from_timeseries(ts):
    return nds2_buffer(ts.name, ts.value, ts.t0.value,
                       ts.sample_rate.value, str(ts.unit))


def nds2_channel(name, sample_rate, unit):
    import nds2
    channel = mock.create_autospec(nds2.channel)
    channel.name = name
    channel.sample_rate = sample_rate
    channel.signal_units = unit
    channel.channel_type = 2
    channel.channel_type_to_string = nds2.channel.channel_type_to_string
    channel.data_type = 8
    for attr, value in inspect.getmembers(
            nds2.channel, predicate=lambda x: isinstance(x, int)):
        setattr(channel, attr, value)
    return channel


def nds2_connection(host='nds.test.gwpy', port=31200, buffers=[], protocol=2):
    import nds2
    NdsConnection = mock.create_autospec(nds2.connection)
    try:
        NdsConnection.get_parameter.return_value = False
    except AttributeError:
        # nds2-client < 0.12 doesn't have {get,set}_parameter
        pass
    NdsConnection.get_host.return_value = host
    NdsConnection.get_port.return_value = int(port)
    NdsConnection.get_protocol.return_value = int(protocol)

    def iterate(start, end, names):
        if not buffers:
            return []
        return [[b for b in buffers if
                 Channel.from_nds2(b.channel).ndsname in names]]

    NdsConnection.iterate = iterate

    def find_channels(name, ctype, dtype, *sample_rate):
        return [b.channel for b in buffers if b.channel.name == name]

    NdsConnection.find_channels = find_channels

    def get_availability(names):
        out = []
        for buff in buffers:
            name = '{0.name},{0.type}'.format(Channel.from_nds2(buff.channel))
            if name not in names:
                segs = []
            else:
                start = buff.gps_seconds + buff.gps_nanoseconds * 1e-9
                end = start + buff.sample_rate * buff.length
                segs = [(start, end)]
            out.append(nds2_availability(name, segs))
        return out

    NdsConnection.get_availability = get_availability

    return NdsConnection


def nds2_availability(name, segments):
    import nds2
    availability = mock.create_autospec(nds2.availability)
    availability.name = name
    availability.simple_list.return_value = list(map(nds2_segment, segments))
    return availability


def nds2_segment(segment):
    import nds2
    nds2seg = mock.create_autospec(nds2.simple_segment)
    nds2seg.gps_start = segment[0]
    nds2seg.gps_stop = segment[1]
    return nds2seg
